{**************************************************************************************************}
{                                                                                                  }
{ VSoft Regular Expressions Wrapper Library                                                        }
{                                                                                                  }
{ The contents of this file are subject to the Mozilla Public License Version 1.1 (the "License"); }
{ you may not use this file except in compliance with the License. You may obtain a copy of the    }
{ License at http://www.mozilla.org/MPL/                                                           }
{                                                                                                  }
{ Software distributed under the License is distributed on an "AS IS" basis, WITHOUT WARRANTY OF   }
{ ANY KIND, either express or implied. See the License for the specific language governing rights  }
{ and limitations under the License.                                                               }
{                                                                                                  }
{ The Original Code is VSoft.RegularExpression.pas.                                                }
{ The Initial Developers of the Original Code are Vincent Parrett, Dick Walker                     }
{ All rights reserved.                                                                             }
{                                                                                                  }
{**************************************************************************************************}
{                                                                                                  }
{ The library is a wrapper around the TPerlRegEx component                                         }
{ TPerlRegEx is available at http://www.regular-expressions.info/delphi.html                       }
{                                                                                                  }
{**************************************************************************************************}


{**************************************************************************************************}
{
  This unit provides a wrapper around TPerlRegEx that attempts to imitate the .NET Regular
  Expression classes in System.Text.RegularExpressions.

  Note that not all of the functionality in RegEx has been implemented, Captures, Split, Match.Result,
  Replaces with MatchEvaluators are not implemented. Accessing Groups by name is also not implemented.

  This unit is designed for Delphi 2009 ONLY. It makes use of D2009 specific features such as
  Generics.

  This unit has an accompanying DUnit Test file : TestVSoftRegularExpression.pas
  -------------------------------------------------------------------------------------------------
   15-03-2009 - Version 0.9
   TESTED!! All functions in this unit have been unit tested. If you add any functions, please add
   unit tests too.
}
{**************************************************************************************************}


unit VSoft.RegularExpression;

interface

type
  RegExOption = (None, IgnoreCase, MultiLine, ExplicitCapture, Compiled, SingleLine, IgnorePatternSpace);
  RegExOptions = set of RegExOption;

  ICapture = interface
    ['{A231A374-C899-4E6D-B5F5-5B23AD19E15E}']
    function GetIndex : integer;
    function GetLength : integer;
    function GetValue : string;

    property Index : integer read GetIndex;
    property Length : integer read GetLength;
    property Value : string read GetValue;
  end;

  (* Not implemented yet...
    ICaptureCollection = interface
    ['{B9295934-2CBC-4C90-AA36-17BC4A1AD1ED}']
    function GetCount : integer;
    function GetItem(index : integer) : ICapture;

    property Count : integer read GetCount;
    property Item[index : integer] : ICapture read GetItem;
  end;*)

  IGroup = interface(ICapture)
  ['{17CF271E-DCAD-4DB9-BDCB-081E2C99D9FC}']
//    function GetCaptures : ICaptureCollection;
    function GetSuccess : boolean;

//    property Captures : ICaptureCollection read GetCaptures;
    property Success : boolean read GetSuccess;
  end;

  IGroupCollection = interface
  ['{5A6653EB-3DAB-4A3D-AA51-4550F51C8C4D}']
    function GetCount : integer;
    function GetItem(index : integer) : IGroup;
    property Count : integer read GetCount;
    property Item[index : integer] : IGroup read GetItem;
  end;

  IMatch = interface(IGroup)
  ['{CD6DF6B3-469C-4C76-A2FE-6871B168B824}']
    function GetGroups : IGroupCollection;
    function NextMatch : IMatch;
    function Result(const replacement : string) : string;
    property Groups : IGroupCollection read GetGroups;
  end;

  MatchEvaluator = reference to function(const m : IMatch) : string;

  IMatchCollection = interface
    ['{8E5660CB-E98A-4E54-BD9E-E643A2EF9B78}']
    function GetCount : integer;
    function GetMatch(index : integer) : IMatch;
    property Count : integer read GetCount;
    property Match[index : integer] : IMatch read GetMatch;
  end;

  IRegularExpression = interface
  ['{5CC9C32E-4F7E-408B-B415-8D72F1E93052}']
    function Match(const input : string) : IMatch;overload;
    function Match(const input : string;const startat : integer) : IMatch;overload;
    function Match(const input : string;const beginning : integer; const length : integer) : IMatch;overload;

    function IsMatch(const input : string) : boolean;overload;
    function IsMatch(const input : string; const startAt : integer) : boolean;overload;

    function Replace(const input : string; const replacement : string) : string;overload;
    function Replace(const input : string; const evaluator : MatchEvaluator) : string;overload;
    function Replace(const input : string; const replacement : string; const count : integer) : string;overload;
    function Replace(const input : string; const evaluator : MatchEvaluator; const count : integer) : string;overload;

    function Matches(const input : string) : IMatchCollection;overload;
    function Matches(const input : string; const startat : integer) : IMatchCollection;overload;

    //TODO - Add Split functions.

  end;

type
  RegEx = class
  public
    class function NewRegularExpression(const pattern : string) : IRegularExpression;  overload;
    class function NewRegularExpression(const pattern : string; const options : RegExOptions) : IRegularExpression;  overload;

    class function IsMatch(const input : string; const pattern : string) : boolean;overload;
    class function IsMatch(const input : string; const pattern : string; const options : RegExOptions) : boolean;overload;

    class function Escape(const str : string; const useWildCards : boolean = false) : string;

    class function Match(const input : string; const pattern : string) : IMatch;overload;
    class function Match(const input : string; const pattern : string; const options : RegExOptions) : IMatch;overload;

    class function Matches(const input : string; const pattern : string) : IMatchCollection;overload;
    class function Matches(const input : string; const pattern : string; const options : RegExOptions) : IMatchCollection;overload;

    class function Replace(const input : string; const pattern : string; const replacement : string) : string;overload;
    class function Replace(const input : string; const pattern : string; const evaluator : MatchEvaluator) : string;overload;
    class function Replace(const input : string; const pattern : string; const replacement : string; const options : RegExOptions) : string;overload;
    class function Replace(const input : string; const pattern : string; const evaluator : MatchEvaluator; const options : RegExOptions) : string;overload;
  end;

implementation

uses
  StrUtils,
  SysUtils,
  classes,
  Generics.Collections,
  PerlRegEx;


function RegExOptionsToPCREOptions(const value : RegExOptions) : TPerlRegExOptions;
begin
  result := [];
  if (IgnoreCase in value) then
    Include(result,preCaseLess);
  if (MultiLine in value) then
    Include(result,preMultiLine);
  if (ExplicitCapture in value) then
    Include(result,preNoAutoCapture);
  if SingleLine in value then
    Include(result,preSingleLine);
  if (IgnorePatternSpace in value) then
    Include(result,preExtended);
end;

type
  IRegExObject = interface
    ['{2CAB6974-C9E8-4709-83FB-23BECE91B538}']
    function RegExObject : TPerlRegEx;
  end;

  //Do not hold references to match or group objects in this object!
  TRegularExpression = class(TInterfacedObject,IRegularExpression,IRegExObject)
  private
    FRegExpr : TPerlRegEx;
    FOptions : RegExOptions;
  protected

    //IRegExObject
    function RegExObject : TPerlRegEx;

    //IRegularExpression
    function Match(const input : string) : IMatch; overload;
    function Match(const input : string;const startat : integer) : IMatch; overload;
    function Match(const input : string;const beginning : integer; const length : integer) : IMatch; overload;

    function IsMatch(const input : string) : boolean; overload;
    function IsMatch(const input : string; const startAt : integer) : boolean; overload;

    function Replace(const input : string; const replacement : string) : string; overload;
    function Replace(const input : string; const evaluator : MatchEvaluator) : string; overload;
    function Replace(const input : string; const replacement : string; const count : integer) : string; overload;
    function Replace(const input : string; const evaluator : MatchEvaluator; const count : integer) : string; overload;

    function Matches(const input : string) : IMatchCollection; overload;
    function Matches(const input : string; const startat : integer) : IMatchCollection; overload;

  public
    constructor Create(const pattern : string); overload;
    constructor Create(const pattern : string; const AOptions : RegExOptions); overload;
    destructor Destroy; override;
  end;

  TCapture = class(TInterfacedObject,ICapture)
  private
    FIndex : Integer;
    FLength : integer;
    FValue : string;
  protected
    function GetIndex: Integer;
    function GetLength: Integer;
    function GetValue: string;
  public
    constructor Create(const AValue : string; const AIndex : integer; const ALength : integer);
  end;

  (*TCaptureCollection = class(TInterfacedObject,ICaptureCollection)
  private
    FList : TList<ICapture>;
  protected
    function GetCount: Integer;
    function GetItem(index: Integer): ICapture;
  public
    constructor Create(const ARegEx : IRegExObject; const AValue : string; const AIndex : integer; const ALength : integer; const ASuccess : boolean);
    destructor Destroy;override;
  end;*)

  TGroup = class(TCapture,IGroup)
  private
    FSuccess : boolean;
    //FCaptures : ICaptureCollection;
  protected
    //function GetCaptures : ICaptureCollection;
    function GetSuccess : boolean;
  public
    constructor Create(const ARegEx : IRegExObject; const AValue : string; const AIndex : integer; const ALength : integer; const ASuccess : boolean);
    destructor Destroy;override;
  end;

  TMatch = class(TGroup,IMatch)
  private
    FGroups : IGroupCollection;
    FRegEx : IRegExObject;
  protected
    function Result(const replacement: string): string;
    //function GetCaptures: ICaptureCollection;
    function NextMatch : IMatch;
    function GetGroups: IGroupCollection;

  public
    constructor Create(const ARegEx : IRegExObject; const AValue : string; const AIndex : integer; const ALength : integer;const ASuccess : boolean);
    destructor Destroy;override;
  end;

  TGroupCollection = class(TInterfacedObject,IGroupCollection)
  private
    FList : TList<IGroup>;
  protected
    function GetCount : integer;
    function GetItem(index : integer) : IGroup;
    property Count : integer read GetCount;
  public
    constructor Create(const ARegEx : IRegExObject; const AValue : string; const AIndex : integer; const ALength : integer;const ASuccess : boolean);
    destructor Destroy;override;
  end;

  TMatchCollection = class(TInterfacedObject,IMatchCollection)
  private
    FList : TList<IMatch>;
  protected
    function GetCount: Integer;
    function GetMatch(index: Integer): IMatch;
  public
    constructor Create(const ARegEx : IRegExObject; const input : string; const AOptions : RegExOptions; const startAt : integer);
    destructor Destroy;override;
  end;

{ TMatch }

constructor TMatch.Create(const ARegEx : IRegExObject; const AValue : string; const AIndex : integer; const ALength : integer;const ASuccess : boolean);
begin
  inherited Create(ARegEx,AValue,AIndex,ALength,ASuccess);
  FRegEx := ARegEx;
  FGroups := TGroupCollection.Create(ARegEx,AValue,AIndex,ALength,ASuccess);
end;

destructor TMatch.Destroy;
begin
  inherited;
end;

(*function TMatch.GetCaptures: ICaptureCollection;
begin

end;*)

function TMatch.GetGroups: IGroupCollection;
begin
  result := FGroups;
end;

function TMatch.NextMatch: IMatch;
var
  bSuccess : boolean;
begin
  bSuccess := FRegEx.RegExObject.MatchAgain;
  if bSuccess then
    result := TMatch.Create(FRegEx,string(FRegEx.RegExObject.Subject),FRegEx.RegExObject.MatchedExpressionOffset,FRegEx.RegExObject.MatchedExpressionLength,bSuccess)
  else
    result := TMatch.Create(FRegEx,string(FRegEx.RegExObject.Subject),0,0,bSuccess)
end;

function TMatch.Result(const replacement: string): string;
//var
//  aRegExpr: TPerlRegEx;
begin
  // DW to be implemented
  raise Exception.Create('not yet implemented');
(*
    aRegExpr := TPerlRegEx.Create(nil);
    aRegExpr.RegEx := UTF8String(replacement);
    aRegExpr.Subject := UTF8String(GetValue);
    if aRegExpr.Match then
      result := string(aRegExpr.SubExpressions[0])
    else
      result := '';*)
end;

{ TRegularExpression }

constructor TRegularExpression.Create(const pattern: string);
begin
  inherited Create;
  FRegExpr := TPerlRegEx.Create(nil);
  FRegExpr.RegEx := UTF8String(pattern);
end;

constructor TRegularExpression.Create(const pattern: string;  const AOptions: RegExOptions);
begin
  inherited Create;
  FRegExpr := TPerlRegEx.Create(nil);
  FRegExpr.Options := RegExOptionsToPCREOptions(AOptions);
  FRegExpr.RegEx := UTF8String(pattern);

  If (Compiled in AOptions) then
    FRegExpr.Compile;
  FOptions := AOptions;
end;

destructor TRegularExpression.Destroy;
begin
  FreeAndNil(FRegExpr);
  inherited;
end;

function TRegularExpression.IsMatch(const input: string; const startAt: integer): boolean;
begin
   FRegExpr.Subject := UTF8String(input);
   FRegExpr.Start := startAt;
   result := FRegExpr.Match;
end;

function TRegularExpression.IsMatch(const input: string): boolean;
begin
   FRegExpr.Subject := UTF8String(input);
   FRegExpr.Start := 1;
   result := FRegExpr.Match;
end;

function TRegularExpression.Match(const input: string): IMatch;
var
  bSuccess : boolean;
begin
  FRegExpr.Subject := UTF8String(input);
  bSuccess := FRegExpr.Match;
  if bSuccess then
    result := TMatch.Create(Self,input,FRegExpr.MatchedExpressionOffset,FRegExpr.MatchedExpressionLength,bSuccess)
  else
    result := TMatch.Create(Self,input,0,0,bSuccess);
end;

function TRegularExpression.Match(const input: string; const startat: integer): IMatch;
var
  bSuccess : boolean;
begin
  FRegExpr.Subject := UTF8String(input);
  FRegExpr.Start := startat;
  bSuccess := FRegExpr.Match;
  result := TMatch.Create(Self,input,FRegExpr.MatchedExpressionOffset,FRegExpr.MatchedExpressionLength,bSuccess);
end;

function TRegularExpression.Match(const input: string; const beginning, length: integer): IMatch;
var
  bSuccess : boolean;
begin
  FRegExpr.Subject := UTF8String(input);
  FRegExpr.Start := beginning;
  FRegExpr.Stop := beginning + length;
  bSuccess := FRegExpr.Match;
  result := TMatch.Create(Self,input,FRegExpr.MatchedExpressionOffset,FRegExpr.MatchedExpressionLength,bSuccess);
end;

function TRegularExpression.RegExObject: TPerlRegEx;
begin
  result := FRegExpr;
end;

function TRegularExpression.Matches(const input: string): IMatchCollection;
begin
  result := TMatchCollection.Create(Self,input,FOptions,1);
end;

function TRegularExpression.Matches(const input: string; const startat: integer): IMatchCollection;
begin
  result := TMatchCollection.Create(Self,input,FOptions,startat);
end;

function TRegularExpression.Replace(const input, replacement: string): string;
begin
  FRegExpr.Subject := UTF8String(input);
  FRegExpr.Replacement := UTF8String(replacement);
  FRegExpr.ReplaceAll;
  result := string(FRegExpr.Subject);
end;

function TRegularExpression.Replace(const input: string; const evaluator: MatchEvaluator): string;
begin
  // DW to be implemented
  raise Exception.Create('not yet implemented');
end;

function TRegularExpression.Replace(const input: string; const evaluator: MatchEvaluator; const count: integer): string;
begin
  // DW to be implemented
  raise Exception.Create('not yet implemented');
end;

function TRegularExpression.Replace(const input, replacement: string; const count: integer): string;
var
  i: integer;
begin
  if count = -1 then
  begin
    result := replace(input, replacement);
    exit;
  end;
  FRegExpr.Subject := UTF8String(input);
  FRegExpr.Replacement := UTF8String(replacement);

  i := 0;
  if FRegExpr.Match then
  begin
    repeat
      FRegExpr.Replace;
      Inc(i)
    until (i = count) or (not FRegExpr.MatchAgain);
  end;
  result := string(FRegExpr.Subject);
end;

{ TCapture }

constructor TCapture.Create(const AValue : string; const AIndex : integer; const ALength : integer);
begin
  FValue := AValue;
  FIndex := AIndex;
  FLength := ALength;
end;

function TCapture.GetIndex: Integer;
begin
  result := FIndex;
end;

function TCapture.GetLength: Integer;
begin
  result := FLength;
end;

function TCapture.GetValue: string;
begin
  //Todo - check if index is 0 or 1 based!
  result := Copy(FValue,FIndex,FLength);
end;

{ TGroup }


{ TGroupCollection }

constructor TGroupCollection.Create(const ARegEx : IRegExObject; const AValue : string; const AIndex : integer; const ALength : integer;const ASuccess : boolean);
var
  i : integer;
  group : IGroup;
begin
  FList := TList<IGroup>.Create;
  /// populate collection;
  if ASuccess then
    for i := 0 to ARegEx.RegExObject.SubExpressionCount do
    begin
      group := TGroup.Create(ARegEx,AValue,ARegEx.RegExObject.SubExpressionOffsets[i],ARegEx.RegExObject.SubExpressionLengths[i],ASuccess);
      FList.Add(group);
    end;
end;

destructor TGroupCollection.Destroy;
begin
  FreeAndNil(FList);
  inherited;
end;

function TGroupCollection.GetCount: integer;
begin
  result := FList.Count;
end;

function TGroupCollection.GetItem(index: integer): IGroup;
begin
  if (index >= 0) and (index < FList.Count) then
    result := FList.Items[index]
  else
    result := nil;
end;

{ TGroup }

constructor TGroup.Create(const ARegEx: IRegExObject; const AValue: string; const AIndex, ALength: integer; const ASuccess : boolean);
begin
  inherited Create(AValue,AIndex,ALength);
  FSuccess := ASuccess;
end;

destructor TGroup.Destroy;
begin
  //FCaptures := nil;
  inherited;
end;

(*function TGroup.GetCaptures: ICaptureCollection;
begin
  result := FCaptures;
end;*)

function TGroup.GetSuccess: boolean;
begin
  result := FSuccess;
end;

class function RegEx.NewRegularExpression(const pattern : string) : IRegularExpression;
begin
  result := TRegularExpression.Create(pattern);
end;

class function RegEx.NewRegularExpression(const pattern : string; const options : RegExOptions) : IRegularExpression;
begin
  result := TRegularExpression.Create(pattern,options);
end;

class function RegEx.IsMatch(const input : string; const pattern : string) : boolean;
var
  regx : IRegularExpression;
  match : IMatch;
begin
  regx := TRegularExpression.Create(pattern);
  match := regx.Match(input);
  result := match.Success;
end;

class function RegEx.IsMatch(const input : string; const pattern : string; const options : RegExOptions) : boolean;
var
  regx : IRegularExpression;
  match : IMatch;
begin
  regx := TRegularExpression.Create(pattern,options);
  match := regx.Match(input);
  result := match.Success;
end;

//Does PCRE do this? No, PCRE only does the 14 chars
class function RegEx.Escape(const str : string; const useWildCards : boolean = false) : string;
const
  Special: array[1..14] of string = ('\', '[', ']', '^', '$', '.', '|', '?', '*', '+', '(', ')', '{', '}');
var
  j: Integer;
begin
  Result := str;
  for j := Low(Special) to High(Special) do
  begin
    Result := StringReplace(Result, Special[j], '\' + Special[j], [rfReplaceAll]);
  end;
  // CRLF becomes \r\n
  Result := StringReplace(Result, #13#10, '\r\n', [rfReplaceAll]);

  if UseWildcards then   // If we're marching wildcards, make them Regex Groups so we can read them back if necessary
  begin
    // Replace all \*s with (.*)
    Result := StringReplace(Result, '\*', '(.*)', [rfReplaceAll]);
    // Replace any \?s with (.)
    Result := StringReplace(Result, '\?', '(.)', [rfReplaceAll]);

    // Wildcards can be escaped as ** or ??
    // Change back any escaped wildcards
    Result := StringReplace(Result, '(.*)(.*)', '\*', [rfReplaceAll]);
    Result := StringReplace(Result, '(.)(.)', '\?', [rfReplaceAll]);
  end;
end;

class function RegEx.Match(const input : string; const pattern : string) : IMatch;
var
  regx : IRegularExpression;
begin
  regx := TRegularExpression.Create(pattern);
  result := regx.Match(input);
end;

class function RegEx.Match(const input : string; const pattern : string; const options : RegExOptions) : IMatch;
var
  regx : IRegularExpression;
begin
  regx := TRegularExpression.Create(pattern,options);
  result := regx.Match(input);
end;

class function RegEx.Matches(const input : string; const pattern : string) : IMatchCollection;
var
  regx : IRegularExpression;
begin
  regx := TRegularExpression.Create(pattern);
  result := regx.Matches(input);
end;

class function RegEx.Matches(const input : string; const pattern : string; const options : RegExOptions) : IMatchCollection;
var
  regx : IRegularExpression;
begin
  regx := TRegularExpression.Create(pattern,options);
  result := regx.Matches(input);
end;

class function RegEx.Replace(const input : string; const pattern : string; const replacement : string) : string;
var
  regx : IRegularExpression;
begin
  regx := TRegularExpression.Create(pattern);
  result := regx.Replace(input,replacement);
end;

class function RegEx.Replace(const input : string; const pattern : string; const evaluator : MatchEvaluator) : string;
var
  regx : IRegularExpression;
begin
  regx := TRegularExpression.Create(pattern);
  result := regx.Replace(input, evaluator);
end;

class function RegEx.Replace(const input : string; const pattern : string; const replacement : string; const options : RegExOptions) : string;
var
  regx : IRegularExpression;
begin
  regx := TRegularExpression.Create(pattern,options);
  result := regx.Replace(input,replacement);
end;

class function RegEx.Replace(const input : string; const pattern : string; const evaluator : MatchEvaluator; const options : RegExOptions) : string;
var
  regx : IRegularExpression;
begin
  regx := TRegularExpression.Create(pattern,options);
  result := regx.Replace(input, evaluator);
end;

{ TCaptureCollection }
(*
constructor TCaptureCollection.Create(const ARegEx: IRegExObject; const AValue : string; const AIndex : integer; const ALength : integer; const ASuccess : boolean);
begin
  FList := TList<ICapture>.Create;
  if ASuccess then
  begin


  end;
end;

destructor TCaptureCollection.Destroy;
begin
  FreeAndNil(FList);
  inherited;
end;

function TCaptureCollection.GetCount: Integer;
begin
  result := FList.Count;
end;

function TCaptureCollection.GetItem(index: Integer): ICapture;
begin
  if (index >= 0) and (index < FList.Count) then
    result := FList.Items[index]
  else
    result := nil;
end;
  *)
{ TMatchCollection }

constructor TMatchCollection.Create(const ARegEx: IRegExObject; const input: string; const AOptions: RegExOptions; const startAt : integer);
var
  res : boolean;
  match : IMatch;
begin
  ARegEx.RegExObject.Subject := UTF8String(input);
  ARegEx.RegExObject.Options := RegExOptionsToPCREOptions(AOptions);
  ARegEx.RegExObject.Start := startAt;
  FList := TList<IMatch>.Create;
  res := ARegEx.RegExObject.Match;
  while res = true do
  begin
    match := TMatch.Create(ARegEx,input,ARegEx.RegExObject.MatchedExpressionOffset,ARegEx.RegExObject.MatchedExpressionLength,res);
    FList.Add(match);
    res := ARegEx.RegExObject.MatchAgain;
  end;
end;

destructor TMatchCollection.Destroy;
begin
  FList.Free;
  inherited;
end;

function TMatchCollection.GetCount: Integer;
begin
  result := FList.Count;
end;

function TMatchCollection.GetMatch(index: Integer): IMatch;
begin
  if (index >= 0) and (index < FList.Count) then
    result := FList.Items[index]
  else
    result := nil;
end;

end.
